from unittest import TestCase

from pythautomata.automata_definitions.weighted_tomitas_grammars import WeightedTomitasGrammars
from pythautomata.utilities.pdfa_operations import get_representative_sample, check_is_minimal
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.automata.wheighted_automaton_definition.probabilistic_deterministic_finite_automaton import \
    ProbabilisticDeterministicFiniteAutomaton as PDFA
from pythautomata.automata.wheighted_automaton_definition.weighted_state import WeightedState
from pythautomata.model_comparators.wfa_tolerance_comparison_strategy import WFAToleranceComparator
binaryAlphabet = Alphabet(frozenset((SymbolStr('0'), SymbolStr('1'))))
zero = binaryAlphabet['0']
one = binaryAlphabet['1']

class TestPDFAOperations(TestCase):

    def test_1(self):
        pdfa = WeightedTomitasGrammars.get_automaton_1()
        sample = get_representative_sample(pdfa, 10)
        self.assertTrue(True)

    def test_2(self):
        pdfa = WeightedTomitasGrammars.get_automaton_2()
        sample = get_representative_sample(pdfa, 20)
        self.assertTrue(True)

    def test_3(self):
        pdfa = WeightedTomitasGrammars.get_automaton_3()
        sample = get_representative_sample(pdfa, 30)
        self.assertTrue(True)

    def test_4(self):
        pdfa = WeightedTomitasGrammars.get_automaton_4()
        sample = get_representative_sample(pdfa, 400)
        self.assertTrue(True)

    def test_5(self):
        pdfa = WeightedTomitasGrammars.get_automaton_5()
        sample = get_representative_sample(pdfa, 0)
        self.assertTrue(True)

    def test_6(self):
        pdfa = WeightedTomitasGrammars.get_automaton_6()
        sample = get_representative_sample(pdfa, 1)
        self.assertTrue(True)

    def test_7(self):
        pdfa = WeightedTomitasGrammars.get_automaton_7()
        self.assertRaises(AssertionError, get_representative_sample, pdfa, -1)
    
    def test_8(self):
        pdfa = WeightedTomitasGrammars.get_automaton_6()
        sample = get_representative_sample(pdfa, 100, 5)
        self.assertTrue(max([len(x) for x in sample])==5)
        self.assertTrue(True)

    def test_pdfa_is_minimal(self):
        for pdfa in WeightedTomitasGrammars.get_all_automata():            
            assert check_is_minimal(pdfa)

    def get_not_minimal_pdfa(self):
        """
        method with a non minimal specification of automaton 1 from the Tomita's Grammars      

        Returns
        -------
        ProbabilisticDeterministicFiniteAutomaton
            non minimal weighted tomita grammar 1
        """
        q0 = WeightedState("q0", 1, 0.05)
        q1 = WeightedState("q1", 0, 0.05)
        q2 = WeightedState("q2", 0, 0.05)
        q3 = WeightedState("q3", 0, 0.05)
        q4 = WeightedState("q4", 0, 0.05)
        q5 = WeightedState("q5", 0, 0.05)
        q6 = WeightedState("q6", 0, 0.05)

        q0.add_transition(zero, q1, 0.665)
        q0.add_transition(one, q0, 0.285)
        q1.add_transition(zero, q2, 0.285)
        q1.add_transition(one, q2, 0.665)
        q2.add_transition(zero, q3, 0.285)
        q2.add_transition(one, q3, 0.665)
        q3.add_transition(zero, q4, 0.285)
        q3.add_transition(one, q4, 0.665)
        q4.add_transition(zero, q5, 0.285)
        q4.add_transition(one, q5, 0.665)
        q5.add_transition(zero, q6, 0.285)
        q5.add_transition(one, q6, 0.665)
        q6.add_transition(zero, q1, 0.285)
        q6.add_transition(one, q1, 0.665)
        

        states = {q0, q1, q2, q3, q4, q5, q6}
        comparator = WFAToleranceComparator()
        return PDFA(binaryAlphabet, states, SymbolStr("$"), comparator, "WeightedTomitas1")

    def test_pdfa_is_not_minimal(self):
        assert not check_is_minimal(self.get_not_minimal_pdfa())


import random
from random import randint

from pythautomata.abstract.finite_automaton import FiniteAutomaton
from pythautomata.base_types.sequence import Sequence

class StatePrefixRandomWalkComparisonStrategy:

    def  __init__(self, number_steps):
        self.steps = number_steps

    def are_equivalent(self, model1, model2):
        return self.get_counterexample_between(model1, model2) is None

    def get_counterexample_between(self, hypothesis: FiniteAutomaton, oracle):
        symbols = list(hypothesis.alphabet.symbols)
        symbols.sort()
        
        for state in hypothesis.states:
            prefix = state.access_string
            if prefix is None:
                prefix = hypothesis.get_access_string(state)

            hypothesis.reset()
            oracle.reset()
            for p in prefix:
                hypothesis.step(p)
                oracle.step(p)

            suffix = Sequence()
            for _ in range(self.steps):
                pos = randint(0, len(symbols) - 1)
                suffix = suffix.append(symbols[pos])

                if hypothesis.step(symbols[pos]) != oracle.step(symbols[pos]):
                    return prefix+suffix
                
        return None



from pythautomata.base_types.sequence import Sequence
from pythautomata.automata.wheighted_automaton_definition.probabilistic_deterministic_finite_automaton import \
    ProbabilisticDeterministicFiniteAutomaton as PDFA
import numpy as np

from pythautomata.utilities.moore_machine_minimizer import MooreMachineMinimizer
from pythautomata.automata.moore_machine_automaton import MooreMachineAutomaton
from pythautomata.base_types.moore_state import MooreState
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.model_comparators.moore_machine_comparison_strategy import MooreMachineComparisonStrategy as MooreMachineComparison
import math


def get_representative_sample(pdfa: PDFA, sample_size: int, max_length: int = None):
    assert (sample_size >= 0)
    sample = list()
    if max_length is None:
        max_length = math.inf
    for i in range(sample_size):
        sample.append(_get_representative_word(pdfa, max_length))
    return sample


def _get_representative_word(pdfa: PDFA, max_length: int):
    word = Sequence()
    first_state = list(filter(lambda x: x.initial_weight ==
                       1, pdfa.weighted_states))[0]
    symbols, weights, next_states = first_state.get_all_symbol_weights()
    next_symbol = np.random.choice(symbols, p=weights)
    total_length = 0
    while next_symbol != pdfa.terminal_symbol and total_length < max_length:
        word += next_symbol
        i = symbols.index(next_symbol)
        next_state = next_states[i]
        symbols, weights, next_states = next_state.get_all_symbol_weights()
        if np.sum(weights)!=1:
            weights = weights/np.sum(weights)
        next_symbol = np.random.choice(symbols, p=weights)
        total_length+=1
    return word


def check_is_minimal(pdfa: PDFA):
    #Build Moore Machine
    moore_states_dict = dict()
    output_symbols = set()
    for state in pdfa.weighted_states:
        _, weights, _ = state.get_all_symbol_weights()
        output_symbol = SymbolStr(str(weights))
        output_symbols.add(output_symbol)
        moore_state = MooreState(state.name, value = output_symbol)
        moore_states_dict[state.name]= moore_state

    output_alphabet = Alphabet(output_symbols)

    for state in pdfa.weighted_states:
        for symbol in pdfa.alphabet.symbols:
            next_state = min(state.next_states_for(symbol))
            moore_states_dict[state.name].add_transition(symbol, moore_states_dict[next_state.name])

    initial_state = moore_states_dict[min(pdfa.initial_states).name]

    moore_machine = MooreMachineAutomaton(input_alphabet=pdfa.alphabet, 
                                          output_alphabet= output_alphabet, 
                                          initial_state= initial_state, 
                                          states = set(moore_states_dict.values()), 
                                          comparator = MooreMachineComparison())

    #Minimize
    minimal_machine = MooreMachineMinimizer(moore_machine).minimize()

    #Compare sizes
    return len(minimal_machine.states) == len(moore_machine.states)


from pythautomata.utilities.sequence_generator import SequenceGenerator
from pythautomata.automata.wheighted_automaton_definition.probabilistic_deterministic_finite_automaton import ProbabilisticDeterministicFiniteAutomaton as PDFA
from pythautomata.base_types.sequence import Sequence
import random

class GuidingPDFASequenceGenerator(SequenceGenerator):
    def __init__(self, pdfa: PDFA, max_seq_lenght: int, random_seed: int = 21, retry = False):
        self.pdfa = pdfa
        self.max_seq_length = max_seq_lenght
        self.retry = retry
        super().__init__(pdfa.alphabet, max_seq_lenght, random_seed)

    def generate_words(self, number_of_words: int):
        result = list()
        for _ in range(number_of_words):
            word = self.generate_single_word(self.max_seq_length)
            result.append(word)
        return result
    
    def _sort_valid_symbol(self, symbols, weights):        
        selected_symbols = []
        for i,w in enumerate(weights):
            if w>0:
                selected_symbols.append(symbols[i])
        next_symbol = random.choice(selected_symbols)
        return next_symbol
    
    def generate_single_word_up_to_length(self, length):
        word = Sequence()
        first_state = list(filter(lambda x: x.initial_weight == 1, self.pdfa.weighted_states))[0]
        symbols, weights, next_states = first_state.get_all_symbol_weights()
        next_symbol = self._sort_valid_symbol(symbols, weights)
        while next_symbol != self.pdfa.terminal_symbol and len(word) <= length:
            word += next_symbol
            i = symbols.index(next_symbol)
            next_state = next_states[i]
            symbols, weights, next_states = next_state.get_all_symbol_weights()
            next_symbol = self._sort_valid_symbol(symbols, weights)
            if next_symbol == self.pdfa.terminal_symbol:
                word += next_symbol            
        is_valid = next_symbol == self.pdfa.terminal_symbol
        return word, is_valid

    def generate_single_word(self, length):
        word, is_valid =  self.generate_single_word_up_to_length(length)
        while not is_valid and self.retry:
            word, is_valid =  self.generate_single_word_up_to_length(length)
        return word


from pythautomata.base_types.symbol import SymbolStr
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.automata.wheighted_automaton_definition.probabilistic_deterministic_finite_automaton import \
    ProbabilisticDeterministicFiniteAutomaton
from pythautomata.automata.wheighted_automaton_definition.weighted_state import WeightedState
from pythautomata.model_comparators.wfa_tolerance_comparison_strategy import WFAToleranceComparator

binaryAlphabet = Alphabet(frozenset((SymbolStr('0'), SymbolStr('1'))))
zero = binaryAlphabet['0']
one = binaryAlphabet['1']


class ZeroTransitionsExamples:
    """
    Class containing weighted (probabilistic) automata with zero valued transitions

    Methods
    -------   
    get_all_automata: list(ProbabilisticDeterministicFiniteAutomaton)
        returns a list containing all weigthed automata defined in this class

    get_automaton_1: ProbabilisticDeterministicFiniteAutomaton
        returns a automaton 1 

    get_automaton_2: ProbabilisticDeterministicFiniteAutomaton
        returns aautomaton 2

    get_automaton_3: ProbabilisticDeterministicFiniteAutomaton
        returns automaton 3 
    """

    @staticmethod
    def get_all_automata():
        """
        Method returning a list of all automata of the class

        Returns
        -------
        list(ProbabilisticDeterministicFiniteAutomaton)
            all automata defined in the class
        """
        return [
            ZeroTransitionsExamples.get_automaton_1(),
            ZeroTransitionsExamples.get_automaton_2(),
            ZeroTransitionsExamples.get_automaton_3(),
        ]

    @staticmethod
    def get_automaton_1():
        """
        method with specification of automaton 1      

        Returns
        -------
        ProbabilisticDeterministicFiniteAutomaton
            automaton 1
        """
        q0 = WeightedState("q0", 1, 0.2)
        q1 = WeightedState("q1", 0, 0.2)

        q0.add_transition(zero, q1, 0)
        q0.add_transition(one, q0, 0.8)
        q1.add_transition(zero, q1, 0.3)
        q1.add_transition(one, q1, 0.5)

        states = {q0, q1}
        comparator = WFAToleranceComparator()
        return ProbabilisticDeterministicFiniteAutomaton(binaryAlphabet, states, SymbolStr("$"), comparator, "ZeroTransitionsExample1")

    @staticmethod
    def get_automaton_2():
        """
        method with specification of automaton 2      

        Returns
        -------
        ProbabilisticDeterministicFiniteAutomaton
            automaton 2
        """
        q0 = WeightedState("q0", 1, 0.2)
        q1 = WeightedState("q1", 0, 0.2)

        q0.add_transition(zero, q1, 0)
        q0.add_transition(one, q0, 0.8)
        q1.add_transition(zero, q1, 0.1)
        q1.add_transition(one, q1, 0.7)

        states = {q0, q1}
        comparator = WFAToleranceComparator()
        return ProbabilisticDeterministicFiniteAutomaton(binaryAlphabet, states, SymbolStr("$"), comparator, "ZeroTransitionsExample2")

    @staticmethod
    def get_automaton_3():
        """
        method with specification of automaton 3      

        Returns
        -------
        ProbabilisticDeterministicFiniteAutomaton
            automaton 3
        """
        q0 = WeightedState("q0", 1, 0.2)
        q1 = WeightedState("q1", 0, 0.2)

        q0.add_transition(zero, q1, 0.05)
        q0.add_transition(one, q0, 0.75)
        q1.add_transition(zero, q1, 0.1)
        q1.add_transition(one, q1, 0.7)

        states = {q0, q1}
        comparator = WFAToleranceComparator()
        return ProbabilisticDeterministicFiniteAutomaton(binaryAlphabet, states, SymbolStr("$"), comparator, "ZeroTransitionsExample3")